------------------------------------------------------------------------
-- More substitutions and relevant lemmas
------------------------------------------------------------------------

-- # import Explicit.Types
-- # import Explicit.Equivalence
-- # import Explicit.Extensional
-- # import Explicit.SubstitutionLemmas
-- # import Explicit.Misc

infixl 80 ↑⋆
infixl 70 /→
infixl 50 ▷

------------------------------------------------------------------------
-- wk⋆

wk⋆ (Γ : Ctxt) : (Γ' : RCtxt Γ) -> Subst Γ (Γ ++ Γ')
wk⋆ (εʳ Γ)    = id Γ
wk⋆ (Γ' ▻ʳ σ) = wk⋆ Γ' ⊙ wk σ

abstract

  wk⋆Cong (Γ¹ : Ctxt) (Γ'¹ : RCtxt Γ¹)
          (Γ² : Ctxt) (Γ'² : RCtxt Γ²)
    : Γ'¹ ≜→ Γ'² -> wk⋆ Γ'¹ ≘ wk⋆ Γ'²
  wk⋆Cong (εʳCong eqΓ)      = idCong eqΓ
  wk⋆Cong (eqΓ' ▻ʳCong eqσ) = compCong (wk⋆Cong eqΓ') (wkCong eqσ)

  singleWk⋆IsWk (Γ : Ctxt)
    : (σ : Ty Γ) -> wk⋆ (singletonʳ σ) ≘ wk σ
  singleWk⋆IsWk σ = idLeftUnit (wk σ)

------------------------------------------------------------------------
-- wk⇒

wk⇒ (Γ, Δ : Ctxt) : Subst Γ Δ -> (σ : Ty Δ) -> Subst Γ (Δ ▻ σ)
wk⇒ ρ σ = ρ ⊙ wk σ

abstract

  wk⇒Cong (Γ¹, Δ¹ : Ctxt) (ρ¹ : Subst Γ¹ Δ¹) (σ¹ : Ty Δ¹)
          (Γ², Δ² : Ctxt) (ρ² : Subst Γ² Δ²) (σ² : Ty Δ²)
    : ρ¹ ≘ ρ² -> σ¹ ≛ σ² -> wk⇒ ρ¹ σ¹ ≘ wk⇒ ρ² σ²
  wk⇒Cong eqρ eqσ = compCong eqρ (wkCong eqσ)

  wk⇒IdIsWk (Γ : Ctxt)
    : (σ : Ty Γ) -> wk⇒ (id Γ) σ ≘ wk σ
  wk⇒IdIsWk σ = idLeftUnit (wk σ)

-----------------------------------------------------------------------
-- wk⇒⋆

wk⇒⋆ (Γ, Δ : Ctxt)
  : Subst Γ Δ -> (Δ' : RCtxt Δ) -> Subst Γ (Δ ++ Δ')
wk⇒⋆ ρ (εʳ _)    = ρ
wk⇒⋆ ρ (Δ' ▻ʳ σ) = wk⇒ (wk⇒⋆ ρ Δ') σ

abstract

  wk⇒⋆Cong (Γ¹, Δ¹ : Ctxt) (ρ¹ : Subst Γ¹ Δ¹) (Δ'¹ : RCtxt Δ¹)
           (Γ², Δ² : Ctxt) (ρ² : Subst Γ² Δ²) (Δ'² : RCtxt Δ²)
    : ρ¹ ≘ ρ² -> Δ'¹ ≜→ Δ'² -> wk⇒⋆ ρ¹ Δ'¹ ≘ wk⇒⋆ ρ² Δ'²
  wk⇒⋆Cong eqρ (εʳCong _)        = eqρ
  wk⇒⋆Cong eqρ (eqΔ' ▻ʳCong eqσ) = wk⇒Cong (wk⇒⋆Cong eqρ eqΔ') eqσ

  wk⇒⋆IsCompWk⋆ (Γ, Δ : Ctxt)
    : (ρ : Subst Γ Δ) -> (Δ' : RCtxt Δ) -> wk⇒⋆ ρ Δ' ≘ ρ ⊙ wk⋆ Δ'
  wk⇒⋆IsCompWk⋆ ρ (εʳ Δ)    = symSubst $ idRightUnit ρ
  wk⇒⋆IsCompWk⋆ ρ (Δ' ▻ʳ σ) =
    compCong (wk⇒⋆IsCompWk⋆ ρ Δ') (reflSubst $ wk σ)
      `transSubst`
    symSubst (compAssoc _ _ _)

-----------------------------------------------------------------------
-- ∅

abstract

  ∅ : (Δ : Ctxt) -> Subst ε Δ
  ∅ ε       = id ε
  ∅ (Δ ▻ σ) = wk⇒ (∅ Δ) σ

  ∅Cong (Δ¹, Δ² : Ctxt)
    : Δ¹ ≜ Δ² -> ∅ Δ¹ ≘ ∅ Δ²
  ∅Cong εCong           = idCong εCong
  ∅Cong (eqΔ ▻Cong eqσ) = wk⇒Cong (∅Cong eqΔ) eqσ

  emptyCtxtSubstIs∅ (Δ : Ctxt)
    : (ρ : Subst ε Δ) -> ρ ≘ ∅ Δ
  emptyCtxtSubstIs∅ ρ = extEqToSubstEq $ extEq⁼ $ \_ v ->
    noVarInEmptyCtxt v

  ∅LeftZero (Δ, X : Ctxt)
    : (ρ : Subst Δ X) -> ∅ Δ ⊙ ρ ≘ ∅ X
  ∅LeftZero {Δ} ρ = emptyCtxtSubstIs∅ (∅ Δ ⊙ ρ)

-----------------------------------------------------------------------
-- ▷

(▷) (Γ, Δ : Ctxt) (σ : Ty Γ)
  : (ρ : Subst Γ Δ) -> Δ ⊢ σ / ρ -> Subst (Γ ▻ σ) Δ
(▷) {Γ} {Δ} {σ} ρ t = (ρ ↑ σ) ⊙ sub t

abstract

  Cong▷ (Γ¹, Δ¹ : Ctxt) (τ¹ : Ty Γ¹)
        (ρ¹ : Subst Γ¹ Δ¹) (t¹ : Δ¹ ⊢ τ¹ / ρ¹)
        (Γ², Δ² : Ctxt) (τ² : Ty Γ²)
        (ρ² : Subst Γ² Δ²) (t² : Δ² ⊢ τ² / ρ²)
    : ρ¹ ≘ ρ² -> τ¹ ≛ τ² -> t¹ ≈⊢ t² -> ρ¹ ▷ t¹ ≘ ρ² ▷ t²
  Cong▷ eqρ eqτ eqt = compCong (liftCong eqρ eqτ) (subCong eqt)

  snocCompIsSnoc (Γ, Δ, X : Ctxt)
    :  (ρ₁ : Subst Γ Δ) -> (ρ₂ : Subst Δ X)
    -> (τ : Ty Γ) -> (t : Δ ⊢ τ / ρ₁)
    -> (eqτ : τ / ρ₁ / ρ₂ ≛ τ / ρ₁ ⊙ ρ₂)
    -> (ρ₁ ▷ t) ⊙ ρ₂ ≘ (ρ₁ ⊙ ρ₂) ▷ (t /⊢ ρ₂ ∷≡⊢ eqτ)
  snocCompIsSnoc ρ₁ ρ₂ τ t eqτ =
    symSubst (compAssoc _ _ _)
      `transSubst`
    compCong (reflSubst _) (subCommutes _ _)
      `transSubst`
    compAssoc _ _ _
      `transSubst`
    symSubst (compCong (compLiftIsComp ρ₁ ρ₂ τ)
                       (subCong $ castTmEq (t /⊢ ρ₂) eqτ))

  wkSnocIsSnocWk (Γ, Δ : Ctxt)
    :  (ρ : Subst Γ Δ) -> (τ : Ty Γ) -> (t : Δ ⊢ τ / ρ) -> (σ : Ty Δ)
    -> (eqτ : τ / ρ / wk σ ≛ τ / wk⇒ ρ σ)
    -> wk⇒ (ρ ▷ t) σ ≘ wk⇒ ρ σ ▷ (t /⊢ wk σ ∷≡⊢ eqτ)
  wkSnocIsSnocWk ρ τ t σ eqτ = snocCompIsSnoc ρ (wk σ) τ t eqτ

  compWkSnocIsLift (Γ, Δ : Ctxt)
    :  (ρ : Subst Γ Δ) -> (σ : Ty Γ)
    -> (eqτ : σ / ρ / wk (σ / ρ) ≛ σ / ρ ⊙ wk (σ / ρ))
    -> (ρ ⊙ wk (σ / ρ)) ▷ var (vz (σ / ρ) ∷≡∋ eqτ) ≘ ρ ↑ σ
  compWkSnocIsLift ρ σ eqτ =
    compCong (compLiftIsComp ρ (wk (σ / ρ)) σ)
             (subCong $ varCong $ castVarEq (vz (σ / ρ)) eqτ)
      `transSubst`
    symSubst (compAssoc _ _ _)
      `transSubst`
    compCong (reflSubst (ρ ↑ σ)) (subVzWithVz _)
      `transSubst`
    idRightUnit (ρ ↑ σ)

  idSnocIsSub (Γ : Ctxt) (τ : Ty Γ)
    :  (t : Γ ⊢ τ) -> (eqτ : τ ≛ τ / id Γ)
    -> id Γ ▷ (t ∷≡⊢ eqτ) ≘ sub t
  idSnocIsSub {Γ} {τ} t eqτ =
    compCong (idLiftIsId τ) (subCong $ castTmEq t eqτ)
      `transSubst`
    idLeftUnit (sub t)

  substVzSub' (Γ, Δ : Ctxt)
    :  (σ : Ty Γ) -> (ρ : Subst Γ Δ) -> (t : Δ ⊢ σ / ρ)
    -> var (vz σ) /⊢ (ρ ▷ t) ≈⊢ t
  substVzSub' σ ρ t =
    compSplitsTm (var (vz σ)) (ρ ↑ σ) (sub t)
      `transTm`
    substTmCong (substVzLift σ ρ) (reflSubst $ sub t)
      `transTm`
    substVzSub (σ / ρ) t

  substVsSub' (Γ, Δ : Ctxt) (τ : Ty Γ)
    :  (v : Γ ∋ τ) -> (σ : Ty Γ) -> (ρ : Subst Γ Δ) -> (t : Δ ⊢ σ / ρ)
    -> var (vs v σ) /⊢ (ρ ▷ t) ≈⊢ var v /⊢ ρ
  substVsSub' v σ ρ t =
    compSplitsTm (var (vs v σ)) (ρ ↑ σ) (sub t)
      `transTm`
    substTmCong (substVsLift v σ ρ) (reflSubst $ sub t)
      `transTm`
    wkSubVanishesTm (var v /⊢ ρ) t

  wkSub'Vanishes (Γ, Δ : Ctxt)
    :  (σ : Ty Γ) -> (ρ : Subst Γ Δ) -> (t : Δ ⊢ σ / ρ)
    -> wk σ ⊙ (ρ ▷ t) ≘ ρ
  wkSub'Vanishes σ ρ t =
    compAssoc (wk σ) (ρ ↑ σ) (sub t)
      `transSubst`
    compCong (wkCommutes σ ρ) (reflSubst _)
      `transSubst`
    symSubst (compAssoc _ _ _)
      `transSubst`
    compCong (reflSubst _) (wkSubVanishes _)
      `transSubst`
    idRightUnit _

  wkSub'VanishesTy (Γ, Δ : Ctxt)
    :  (τ, σ : Ty Γ) -> (ρ : Subst Γ Δ) -> (t : Δ ⊢ σ / ρ)
    -> τ / wk σ / (ρ ▷ t) ≛ τ / ρ
  wkSub'VanishesTy τ σ ρ t =
    symTy (compSplitsTy τ (wk σ) (ρ ▷ t))
      `transTy`
    substTyCong (reflTy τ) (wkSub'Vanishes σ ρ t)

  initIsInit (Γ, Δ : Ctxt)
    :  (τ : Ty Γ) -> (ρ : Subst Γ Δ) -> (t : Δ ⊢ τ / ρ)
    -> init (ρ ▷ t) ≘ ρ
  initIsInit τ ρ t = wkSub'Vanishes τ ρ t

  lastIsLast (Γ, Δ : Ctxt)
    :  (τ : Ty Γ) -> (ρ : Subst Γ Δ) -> (t : Δ ⊢ τ / ρ)
    -> last (ρ ▷ t) ≈⊢ t
  lastIsLast τ ρ t = castTmEqˡ _ $ substVzSub' _ ρ t

  nonEmptyCtxtSubstIs▷ (Γ, Δ : Ctxt) (τ : Ty Γ)
    : (ρ : Subst (Γ ▻ τ) Δ) -> ρ ≘ init ρ ▷ last ρ
  nonEmptyCtxtSubstIs▷ {Γ} {Δ} {τ} ρ =
    symSubst (idLeftUnit _)
      `transSubst`
    symSubst (compCong (subVzWithVz _) (reflSubst _))
      `transSubst`
    snocCompIsSnoc (wk τ) ρ τ (var (vz τ))
                   (symTy (compSplitsTy τ (wk τ) ρ))

-----------------------------------------------------------------------
-- ↑⋆ and /→

mutual

  -- Applies a substitution to an RCtxt.

  (/→) (Γ, Δ : Ctxt)
    : RCtxt Γ -> Subst Γ Δ -> RCtxt Δ
  (εʳ _)    /→ ρ = εʳ _
  (Γ' ▻ʳ τ) /→ ρ = Γ' /→ ρ ▻ʳ τ / (ρ ↑⋆ Γ')

  -- ↑⋆ performs multiple liftings at one time.

  (↑⋆) (Γ, Δ : Ctxt)
    :  (ρ : Subst Γ Δ) -> (Γ' : RCtxt Γ)
    -> Subst (Γ ++ Γ') (Δ ++ Γ' /→ ρ)
  ρ ↑⋆ (εʳ _)    = ρ
  ρ ↑⋆ (Γ' ▻ʳ τ) = (ρ ↑⋆ Γ') ↑ τ

abstract

  liftʳCong (Γ¹, Δ¹ : Ctxt) (Γ'¹ : RCtxt Γ¹) (ρ¹ : Subst Γ¹ Δ¹)
            (Γ², Δ² : Ctxt) (Γ'² : RCtxt Γ²) (ρ² : Subst Γ² Δ²)
    : ρ¹ ≘ ρ² -> Γ'¹ ≜→ Γ'² -> ρ¹ ↑⋆ Γ'¹ ≘ ρ² ↑⋆ Γ'²
  liftʳCong eqρ (εʳCong eqΓ)      = eqρ
  liftʳCong eqρ (eqΓ' ▻ʳCong eqτ) = liftCong (liftʳCong eqρ eqΓ') eqτ

  substRCtxtCong (Γ¹, Δ¹ : Ctxt) (Γ'¹ : RCtxt Γ¹) (ρ¹ : Subst Γ¹ Δ¹)
                 (Γ², Δ² : Ctxt) (Γ'² : RCtxt Γ²) (ρ² : Subst Γ² Δ²)
    : Γ'¹ ≜→ Γ'² -> ρ¹ ≘ ρ² -> Γ'¹ /→ ρ¹ ≜→ Γ'² /→ ρ²
  substRCtxtCong (εʳCong eqΓ)      eqρ = εʳCong (substEqCtxtEqRight eqρ)
  substRCtxtCong (eqΓ' ▻ʳCong eqτ) eqρ =
    substRCtxtCong eqΓ' eqρ ▻ʳCong substTyCong eqτ (liftʳCong eqρ eqΓ')

abstract

  idLiftʳIsId (Γ : Ctxt)
    : (Γ' : RCtxt Γ) -> id Γ ↑⋆ Γ' ≘ id (Γ ++ Γ')
  idLiftʳIsId (εʳ Γ)    = idCong (reflCtxt Γ)
  idLiftʳIsId (Γ' ▻ʳ τ) =
    liftCong (idLiftʳIsId Γ') (reflTy τ)
      `transSubst`
    idLiftIsId τ

  idVanishesTyʳ (Γ : Ctxt)
    : (Γ' : RCtxt Γ) -> (τ : Ty (Γ ++ Γ')) -> τ / id Γ ↑⋆ Γ' ≛ τ
  idVanishesTyʳ Γ' τ =
    substTyCong (reflTy τ) (idLiftʳIsId Γ')
      `transTy`
    idVanishesTy τ

  idVanishesRCtxt (Γ : Ctxt)
    : (Γ' : RCtxt Γ) -> Γ' /→ id Γ ≜→ Γ'
  idVanishesRCtxt (εʳ Γ)    = εʳCong $ reflCtxt Γ
  idVanishesRCtxt (Γ' ▻ʳ τ) =
    idVanishesRCtxt Γ' ▻ʳCong idVanishesTyʳ Γ' τ

abstract

  compLiftʳIsComp (Γ, Δ, X : Ctxt)
    :  (ρ₁ : Subst Γ Δ) -> (ρ₂ : Subst Δ X) -> (Γ' : RCtxt Γ)
    -> ρ₁ ⊙ ρ₂ ↑⋆ Γ' ≘ (ρ₁ ↑⋆ Γ') ⊙ (ρ₂ ↑⋆ (Γ' /→ ρ₁))
  compLiftʳIsComp {Γ} {Δ} {X} ρ₁ ρ₂ (εʳ Γ)    = reflSubst (ρ₁ ⊙ ρ₂)
  compLiftʳIsComp {Γ} {Δ} {X} ρ₁ ρ₂ (Γ' ▻ʳ σ) =
    liftCong (compLiftʳIsComp ρ₁ ρ₂ Γ') (reflTy σ)
      `transSubst`
    compLiftIsComp (ρ₁ ↑⋆ Γ') (ρ₂ ↑⋆ (Γ' /→ ρ₁)) σ

  compSplitsTyʳ (Γ, Δ, X : Ctxt)
    :  (Γ' : RCtxt Γ) -> (τ : Ty (Γ ++ Γ'))
    -> (ρ₁ : Subst Γ Δ) -> (ρ₂ : Subst Δ X)
    -> τ / ρ₁ ⊙ ρ₂ ↑⋆ Γ' ≛ τ / ρ₁ ↑⋆ Γ' / ρ₂ ↑⋆ (Γ' /→ ρ₁)
  compSplitsTyʳ Γ' τ ρ₁ ρ₂ =
    substTyCong (reflTy τ) (compLiftʳIsComp ρ₁ ρ₂ Γ')
      `transTy`
    compSplitsTy τ (ρ₁ ↑⋆ Γ') (ρ₂ ↑⋆ (Γ' /→ ρ₁))

  compSplitsRCtxt (Γ, Δ, X : Ctxt)
    :  (Γ' : RCtxt Γ) -> (ρ₁ : Subst Γ Δ) -> (ρ₂ : Subst Δ X)
    -> Γ' /→ ρ₁ ⊙ ρ₂ ≜→ Γ' /→ ρ₁ /→ ρ₂
  compSplitsRCtxt {Γ} {Δ} {X} (εʳ Γ)    ρ₁ ρ₂ = εʳCong (reflCtxt X)
  compSplitsRCtxt             (Γ' ▻ʳ τ) ρ₁ ρ₂ =
    compSplitsRCtxt Γ' ρ₁ ρ₂ ▻ʳCong compSplitsTyʳ Γ' τ ρ₁ ρ₂
