------------------------------------------------------------------------
-- Equalities used for the Implicit types
------------------------------------------------------------------------

-- The equalities are indexed on equality kinds; the β/η axioms may
-- only be used if the index is βη.

-- # import Explicit
-- # import Implicit.Terms
-- # import Implicit.Substitutions
-- # import EqKind
-- # import TmVar

infix 40 ≈⊢⁻, ≛⁻, ≘⁻

------------------------------------------------------------------------
-- Tm⁻ equality

mutual

  Tm⁻Eq (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹) (tv¹ : TmVar)
        (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²) (tv² : TmVar)
    : EqKind -> Tm⁻ tv¹ t¹ -> Tm⁻ tv² t² -> Set
  Tm⁻Eq = Tm⁻Eq' _ _ _ _ _ _ _ _

  data Tm⁻Eq'
   :  (Γ¹ : Ctxt) -> (τ¹ : Ty Γ¹) -> (t¹ : Γ¹ ⊢ τ¹) -> (tv¹ : TmVar)
   -> (Γ² : Ctxt) -> (τ² : Ty Γ²) -> (t² : Γ² ⊢ τ²) -> (tv² : TmVar)
   -> EqKind -> Tm⁻ tv¹ t¹ -> Tm⁻ tv² t² -> Set where

    -- Congruence.

    var⁻Cong (k : EqKind)
             (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (v¹ : Γ¹ ∋ τ¹) (tv¹ : TmVar)
             (Γ² : Ctxt) (τ² : Ty Γ²) (v² : Γ² ∋ τ²) (tv² : TmVar)
      : v¹ ≈∋ v² -> Tm⁻Eq k (var⁻ tv¹ v¹) (var⁻ tv² v²)
    λ⁻Cong (k : EqKind)
           (Γ¹ : Ctxt) (τ₁¹ : Ty Γ¹) (τ₂¹ : Ty (Γ¹ ▻ τ₁¹)) (tv¹ : TmVar)
           (t¹ : Γ¹ ▻ τ₁¹ ⊢ τ₂¹) (τ⁻₁¹ : Ty⁻ τ₁¹) (t⁻¹ : Tm⁻ tv¹ t¹)
           (Γ² : Ctxt) (τ₁² : Ty Γ²) (τ₂² : Ty (Γ² ▻ τ₁²)) (tv² : TmVar)
           (t² : Γ² ▻ τ₁² ⊢ τ₂²) (τ⁻₁² : Ty⁻ τ₁²) (t⁻² : Tm⁻ tv² t²)
      : Tm⁻Eq k t⁻¹ t⁻² -> Tm⁻Eq k (λ⁻ τ⁻₁¹ t⁻¹) (λ⁻ τ⁻₁² t⁻²)
    app⁻Cong (k : EqKind)
             (Γ¹ : Ctxt) (τ₁¹ : Ty Γ¹) (τ₂¹ : Ty (Γ¹ ▻ τ₁¹))
             (t₁¹ : Γ¹ ⊢ Π τ₁¹ τ₂¹) (t₂¹ : Γ¹ ⊢ τ₁¹)
             (tv₁¹, tv₂¹ : TmVar)
             (t⁻₁¹ : Tm⁻ tv₁¹ t₁¹) (t⁻₂¹ : Tm⁻ tv₂¹ t₂¹)
             (Γ² : Ctxt) (τ₁² : Ty Γ²) (τ₂² : Ty (Γ² ▻ τ₁²))
             (t₁² : Γ² ⊢ Π τ₁² τ₂²) (t₂² : Γ² ⊢ τ₁²)
             (tv₁², tv₂² : TmVar)
             (t⁻₁² : Tm⁻ tv₁² t₁²) (t⁻₂² : Tm⁻ tv₂² t₂²)
      :  Tm⁻Eq k t⁻₁¹ t⁻₁² -> Tm⁻Eq k t⁻₂¹ t⁻₂²
      -> Tm⁻Eq k (t⁻₁¹ @⁻ t⁻₂¹) (t⁻₁² @⁻ t⁻₂²)

    -- Casts.

    castTm⁻Eqˡ (k : EqKind)
               (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹) (tv¹ : TmVar)
               (t⁻¹ : Tm⁻ tv¹ t¹)
               (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²) (tv² : TmVar)
               (t⁻² : Tm⁻ tv² t²)
                           (τ³ : Ty Γ¹) (t³ : Γ¹ ⊢ τ³)
      : (eqt : t¹ ≈⊢ t³) -> Tm⁻Eq k t⁻¹ t⁻² -> Tm⁻Eq k (t⁻¹ ∷≡⊢⁻ eqt) t⁻²
    castTm⁻Eqʳ (k : EqKind)
               (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹) (tv¹ : TmVar)
               (t⁻¹ : Tm⁻ tv¹ t¹)
               (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²) (tv² : TmVar)
               (t⁻² : Tm⁻ tv² t²)
                           (τ³ : Ty Γ²) (t³ : Γ² ⊢ τ³)
      : (eqt : t² ≈⊢ t³) -> Tm⁻Eq k t⁻¹ t⁻² -> Tm⁻Eq k t⁻¹ (t⁻² ∷≡⊢⁻ eqt)

    -- Only some of the βη equivalence axioms are postulated here;
    -- more general variants are proved in Implicit.Equivalence.

    symTm⁻βη (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹) (tv¹ : TmVar)
             (t⁻¹ : Tm⁻ tv¹ t¹)
             (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²) (tv² : TmVar)
             (t⁻² : Tm⁻ tv² t²)
      : Tm⁻Eq βη t⁻¹ t⁻² -> Tm⁻Eq βη t⁻² t⁻¹
    transTm⁻βη (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹) (tv¹ : TmVar)
               (t⁻¹ : Tm⁻ tv¹ t¹)
               (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²) (tv² : TmVar)
               (t⁻² : Tm⁻ tv² t²)
               (Γ³ : Ctxt) (τ³ : Ty Γ³) (t³ : Γ³ ⊢ τ³) (tv³ : TmVar)
               (t⁻³ : Tm⁻ tv³ t³)
      : Tm⁻Eq βη t⁻¹ t⁻² -> Tm⁻Eq βη t⁻² t⁻³ -> Tm⁻Eq βη t⁻¹ t⁻³

    -- Evaluation.

    β⁻ (tv₁, tv₂ : TmVar)
       (Γ : Ctxt) (τ₁ : Ty Γ) (τ₂ : Ty (Γ ▻ τ₁)) (τ⁻₁ : Ty⁻ τ₁)
       (t₁ : Γ ▻ τ₁ ⊢ τ₂) (t₂ : Γ ⊢ τ₁)
      :  (t⁻₁ : Tm⁻ tv₁ t₁) -> (t⁻₂ : Tm⁻ tv₂ t₂)
      -> Tm⁻Eq βη (λ⁻ τ⁻₁ t⁻₁ @⁻ t⁻₂) (t⁻₁ /⊢⁻ sub⁻ t⁻₂)

    -- η equality.

    η⁻ (tv₁, tv₂, tv₃ : TmVar) (Γ : Ctxt)
       (τ₁ : Ty Γ) (τ₂ : Ty (Γ ▻ τ₁)) (t : Γ ⊢ Π τ₁ τ₂) (τ⁻₁ : Ty⁻ τ₁)
      :  (t⁻ : Tm⁻ tv₁ t)
      -> Tm⁻Eq βη (λ⁻ τ⁻₁ ((t⁻ /⊢⁻ wk⁻ tv₂ τ₁) @⁻ var⁻ tv₃ (vz τ₁))) t⁻

(≈⊢⁻) (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹) (tv¹ : TmVar)
      (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²) (tv² : TmVar)
  : Tm⁻ tv¹ t¹ -> Tm⁻ tv² t² -> Set
(≈⊢⁻) = Tm⁻Eq βη

------------------------------------------------------------------------
-- Ty⁻ equality

mutual

  (≛⁻) (Γ¹ : Ctxt) (τ¹ : Ty Γ¹)
       (Γ² : Ctxt) (τ² : Ty Γ²)
    : Ty⁻ τ¹ -> Ty⁻ τ² -> Set
  (≛⁻) = Ty⁻Eq _ _ _ _

  data Ty⁻Eq  :  (Γ¹ : Ctxt) -> (τ¹ : Ty Γ¹)
              -> (Γ² : Ctxt) -> (τ² : Ty Γ²)
              -> Ty⁻ τ¹ -> Ty⁻ τ² -> Set where

    -- Congruence.

    ✶⁻Cong (Γ¹, Γ² : Ctxt)
      : Γ¹ ≜ Γ² -> ✶⁻ Γ¹ ≛⁻ ✶⁻ Γ²
    Π⁻Cong (Γ¹ : Ctxt) (τ₁¹ : Ty Γ¹) (τ₂¹ : Ty (Γ¹ ▻ τ₁¹))
           (τ⁻₁¹ : Ty⁻ τ₁¹) (τ⁻₂¹ : Ty⁻ τ₂¹)
           (Γ² : Ctxt) (τ₁² : Ty Γ²) (τ₂² : Ty (Γ² ▻ τ₁²))
           (τ⁻₁² : Ty⁻ τ₁²) (τ⁻₂² : Ty⁻ τ₂²)
     : τ⁻₁¹ ≛⁻ τ⁻₁² -> τ⁻₂¹ ≛⁻ τ⁻₂² -> Π⁻ τ⁻₁¹ τ⁻₂¹ ≛⁻ Π⁻ τ⁻₁² τ⁻₂²
    El⁻Cong
      (Γ¹ : Ctxt) (t¹ : Γ¹ ⊢ ✶ Γ¹) (tv¹ : TmVar) (t⁻¹ : Tm⁻ tv¹ t¹)
      (Γ² : Ctxt) (t² : Γ² ⊢ ✶ Γ²) (tv² : TmVar) (t⁻² : Tm⁻ tv² t²)
      : t⁻¹ ≈⊢⁻ t⁻² -> El⁻ t⁻¹ ≛⁻ El⁻ t⁻²

    -- Casts.

    castTy⁻Eqˡ (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (τ⁻¹ : Ty⁻ τ¹)
               (Γ² : Ctxt) (τ² : Ty Γ²) (τ⁻² : Ty⁻ τ²)
                           (τ³ : Ty Γ¹)
      : (eqτ : τ¹ ≛ τ³) -> τ⁻¹ ≛⁻ τ⁻² -> τ⁻¹ ∷≡✶⁻ eqτ ≛⁻ τ⁻²
    castTy⁻Eqʳ (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (τ⁻¹ : Ty⁻ τ¹)
               (Γ² : Ctxt) (τ² : Ty Γ²) (τ⁻² : Ty⁻ τ²)
                           (τ³ : Ty Γ²)
      : (eqτ : τ² ≛ τ³) -> τ⁻¹ ≛⁻ τ⁻² -> τ⁻¹ ≛⁻ τ⁻² ∷≡✶⁻ eqτ

------------------------------------------------------------------------
-- Subst⁻ equality

ExtEq⁻ (tv¹ : TmVar) (Γ¹, Δ¹ : Ctxt) (ρ¹ : Subst Γ¹ Δ¹)
       (tv² : TmVar) (Γ², Δ² : Ctxt) (ρ² : Subst Γ² Δ²)
  : EqKind -> Subst⁻ tv¹ ρ¹ -> Subst⁻ tv² ρ² -> Set
ExtEq⁻ {tv¹} {Γ¹} {Δ¹} {ρ¹} {tv²} {Γ²} {Δ²} {ρ²} k ρ⁻¹ ρ⁻² =
     (σ¹ : Ty Γ¹) -> (v¹ : Γ¹ ∋ σ¹)
  -> (σ² : Ty Γ²) -> (v² : Γ² ∋ σ²)
  -> v¹ ≈∋ v² -> Tm⁻Eq k (v¹ /∋⁻ ρ⁻¹) (v² /∋⁻ ρ⁻²)

data Subst⁻Eq {tv¹ : TmVar} {Γ¹, Δ¹ : Ctxt} {ρ¹ : Subst Γ¹ Δ¹}
              {tv² : TmVar} {Γ², Δ² : Ctxt} {ρ² : Subst Γ² Δ²}
              (k : EqKind) (ρ⁻¹ : Subst⁻ tv¹ ρ¹) (ρ⁻² : Subst⁻ tv² ρ²)
              : Set where
  extEq⁻ : Γ¹ ≜ Γ² -> Δ¹ ≜ Δ² -> ExtEq⁻ k ρ⁻¹ ρ⁻² -> Subst⁻Eq k ρ⁻¹ ρ⁻²

(≘⁻) (tv¹ : TmVar) (Γ¹, Δ¹ : Ctxt) (ρ¹ : Subst Γ¹ Δ¹)
     (tv² : TmVar) (Γ², Δ² : Ctxt) (ρ² : Subst Γ² Δ²)
  : (ρ⁻¹ : Subst⁻ tv¹ ρ¹) -> (ρ⁻² : Subst⁻ tv² ρ²) -> Set
(≘⁻) = Subst⁻Eq βη

Implicit-Equality : Type
Implicit-Equality = time-plugin(Implicit-Equality, Set)
