------------------------------------------------------------------------
-- Weakening congruences
------------------------------------------------------------------------

-- These data types are used to parametrise a proof mentioning a
-- weakening w on a proof showing that w is a congruence.

-- # import Explicit
-- # import Implicit.Terms
-- # import Implicit.Substitutions
-- # import Implicit.Equality
-- # import Implicit.Equivalence
-- # import EqKind
-- # import TmVar

data WeakeningCong {tv¹, tv² : TmVar}
       (k : EqKind) (w¹ : Weakening tv¹) (w² : Weakening tv²)
       : Set where
  weakeningCong
    :  (   (Γ¹ : Ctxt) -> (τ¹, σ¹ : Ty Γ¹)
        -> (t¹ : Γ¹ ⊢ τ¹) -> (t⁻¹ : Tm⁻ tv¹ t¹)
        -> (Γ² : Ctxt) -> (τ², σ² : Ty Γ²)
        -> (t² : Γ² ⊢ τ²) -> (t⁻² : Tm⁻ tv² t²)
        -> Tm⁻Eq k t⁻¹ t⁻² -> σ¹ ≛ σ²
        -> Tm⁻Eq k (applyWeakening w¹ t⁻¹ σ¹)
                   (applyWeakening w² t⁻² σ²))
    -> WeakeningCong k w¹ w²

abstract

  applyWeakeningCong (k : EqKind)
                     (tv¹ : TmVar) (w¹ : Weakening tv¹)
                     (Γ¹ : Ctxt) (τ¹, σ¹ : Ty Γ¹)
                     (t¹ : Γ¹ ⊢ τ¹) (t⁻¹ : Tm⁻ tv¹ t¹)
                     (tv² : TmVar) (w² : Weakening tv²)
                     (Γ² : Ctxt) (τ², σ² : Ty Γ²)
                     (t² : Γ² ⊢ τ²) (t⁻² : Tm⁻ tv² t²)
    :  WeakeningCong k w¹ w² -> Tm⁻Eq k t⁻¹ t⁻² -> σ¹ ≛ σ²
    -> Tm⁻Eq k (applyWeakening w¹ t⁻¹ σ¹)
               (applyWeakening w² t⁻² σ²)
  applyWeakeningCong (weakeningCong f) = f _ _ _ _ _ _ _ _ _ _

  swapWCong (k : EqKind)
            (tv¹ : TmVar) (w¹ : Weakening tv¹)
            (tv² : TmVar) (w² : Weakening tv²)
    : WeakeningCong k w¹ w² -> WeakeningCong k w² w¹
  swapWCong wCong = weakeningCong $
    \_ _ _ _ _ _ _ _ _ _ eqt⁻ eqσ ->
      symTm⁻ $ applyWeakeningCong wCong (symTm⁻ eqt⁻) (symTy eqσ)

WeakeningCong₁ (tv : TmVar) : EqKind -> Weakening tv -> Set
WeakeningCong₁ k w = WeakeningCong k w w

data WeakeningCong₁₂₁ {tv¹, tv² : TmVar}
       (k : EqKind) (w¹ : Weakening tv¹) (w² : Weakening tv²)
       : Set where
  weakeningCong₁₂₁
    :  WeakeningCong₁ k w¹
    -> WeakeningCong k w¹ w²
    -> WeakeningCong₁ k w²
    -> WeakeningCong₁₂₁ k w¹ w²

abstract

  swapWCong₁₂₁ (k : EqKind)
               (tv¹ : TmVar) (w¹ : Weakening tv¹)
               (tv² : TmVar) (w² : Weakening tv²)
    : WeakeningCong₁₂₁ k w¹ w² -> WeakeningCong₁₂₁ k w² w¹
  swapWCong₁₂₁ (weakeningCong₁₂₁ wCong¹ wCong¹² wCong²) =
    weakeningCong₁₂₁ wCong² (swapWCong wCong¹²) wCong¹

  wCong₁To₁₂₁ (k : EqKind) (tv : TmVar) (w : Weakening tv)
    : WeakeningCong₁ k w -> WeakeningCong₁₂₁ k w w
  wCong₁To₁₂₁ wCong = weakeningCong₁₂₁ wCong wCong wCong

  wCong₁₂₁ToL (k : EqKind)
              (tv¹ : TmVar) (w¹ : Weakening tv¹)
              (tv² : TmVar) (w² : Weakening tv²)
    : WeakeningCong₁₂₁ k w¹ w² -> WeakeningCong₁ k w¹
  wCong₁₂₁ToL (weakeningCong₁₂₁ wCong¹ wCong¹² wCong²) = wCong¹

  wCong₁₂₁ToR (k : EqKind)
              (tv¹ : TmVar) (w¹ : Weakening tv¹)
              (tv² : TmVar) (w² : Weakening tv²)
    : WeakeningCong₁₂₁ k w¹ w² -> WeakeningCong₁ k w²
  wCong₁₂₁ToR (weakeningCong₁₂₁ wCong¹ wCong¹² wCong²) = wCong²

  wCong₁₂₁To₁₁₁ (k : EqKind)
                (tv¹ : TmVar) (w¹ : Weakening tv¹)
                (tv² : TmVar) (w² : Weakening tv²)
    : WeakeningCong₁₂₁ k w¹ w² -> WeakeningCong₁₂₁ k w¹ w¹
  wCong₁₂₁To₁₁₁ wCong = wCong₁To₁₂₁ $ wCong₁₂₁ToL wCong

  wCong₁₂₁To₂₂₂ (k : EqKind)
                (tv¹ : TmVar) (w¹ : Weakening tv¹)
                (tv² : TmVar) (w² : Weakening tv²)
    : WeakeningCong₁₂₁ k w¹ w² -> WeakeningCong₁₂₁ k w² w²
  wCong₁₂₁To₂₂₂ wCong = wCong₁₂₁To₁₁₁ $ swapWCong₁₂₁ wCong

  applyWeakeningCongˡ (k : EqKind)
                      (tv¹ : TmVar) (w¹ : Weakening tv¹)
                      (tv² : TmVar) (w² : Weakening tv²)
                      (Γ¹ : Ctxt) (τ¹, σ¹ : Ty Γ¹)
                      (t¹ : Γ¹ ⊢ τ¹) (t⁻¹ : Tm⁻ tv¹ t¹)
                      (Γ² : Ctxt) (τ², σ² : Ty Γ²)
                      (t² : Γ² ⊢ τ²) (t⁻² : Tm⁻ tv¹ t²)
    :  WeakeningCong₁₂₁ k w¹ w² -> Tm⁻Eq k t⁻¹ t⁻² -> σ¹ ≛ σ²
    -> Tm⁻Eq k (applyWeakening w¹ t⁻¹ σ¹)
               (applyWeakening w¹ t⁻² σ²)
  applyWeakeningCongˡ (weakeningCong₁₂₁ w¹¹ w¹² w²²) =
    applyWeakeningCong w¹¹

  applyWeakeningCongʳ (k : EqKind)
                      (tv¹ : TmVar) (w¹ : Weakening tv¹)
                      (tv² : TmVar) (w² : Weakening tv²)
                      (Γ¹ : Ctxt) (τ¹, σ¹ : Ty Γ¹)
                      (t¹ : Γ¹ ⊢ τ¹) (t⁻¹ : Tm⁻ tv² t¹)
                      (Γ² : Ctxt) (τ², σ² : Ty Γ²)
                      (t² : Γ² ⊢ τ²) (t⁻² : Tm⁻ tv² t²)
    :  WeakeningCong₁₂₁ k w¹ w² -> Tm⁻Eq k t⁻¹ t⁻² -> σ¹ ≛ σ²
    -> Tm⁻Eq k (applyWeakening w² t⁻¹ σ¹)
               (applyWeakening w² t⁻² σ²)
  applyWeakeningCongʳ (weakeningCong₁₂₁ w¹¹ w¹² w²²) =
    applyWeakeningCong w²²

  applyWeakeningCongˡʳ (k : EqKind)
                       (tv¹ : TmVar) (w¹ : Weakening tv¹)
                       (Γ¹ : Ctxt) (τ¹, σ¹ : Ty Γ¹)
                       (t¹ : Γ¹ ⊢ τ¹) (t⁻¹ : Tm⁻ tv¹ t¹)
                       (tv² : TmVar) (w² : Weakening tv²)
                       (Γ² : Ctxt) (τ², σ² : Ty Γ²)
                       (t² : Γ² ⊢ τ²) (t⁻² : Tm⁻ tv² t²)
    :  WeakeningCong₁₂₁ k w¹ w² -> Tm⁻Eq k t⁻¹ t⁻² -> σ¹ ≛ σ²
    -> Tm⁻Eq k (applyWeakening w¹ t⁻¹ σ¹)
               (applyWeakening w² t⁻² σ²)
  applyWeakeningCongˡʳ (weakeningCong₁₂₁ w¹¹ w¹² w²²) =
    applyWeakeningCong w¹²

Implicit-WeakeningCong : Type
Implicit-WeakeningCong = time-plugin(Implicit-WeakeningCong, Set)
