------------------------------------------------------------------------
-- Environments
------------------------------------------------------------------------

-- # import Explicit
-- # import Implicit
-- # import Normalisation.NormalForms
-- # import Normalisation.Values
-- # import TmVar

-- The ˢ index stands for "_s_emantical value".

infixl 70 /∋ˢ
infixl 50 ▷ˢ
infixl 45 ∷≡⇒ˢ, ∷⇒ˢ

------------------------------------------------------------------------
-- Value substitutions (environments)

-- Since Substˢ basically is a copy of Subst⁻ it would be nice if the
-- two definitions could be merged.

mutual

  Substˢ (Γ, Δ : Ctxt) : Subst Γ Δ -> Set
  Substˢ = Substˢ' _ _

  data Substˢ' :  (Γ, Δ : Ctxt) -> Subst Γ Δ -> Set where
    ∅ˢ : (Δ : Ctxt) -> Substˢ (∅ Δ)
    (▷ˢ) (Γ, Δ : Ctxt) (σ : Ty Γ) (ρ : Subst Γ Δ) (t : Δ ⊢ σ / ρ)
      : Substˢ ρ -> (v : Val t) -> Substˢ (ρ ▷ t)
    (∷≡⇒ˢ) (Γ, Δ : Ctxt) (ρ¹ : Subst Γ Δ) (ρ² : Subst Γ Δ)
      : Substˢ ρ¹ -> ρ¹ ≘ ρ² -> Substˢ ρ²

substˢToSubst⁻ (Γ, Δ : Ctxt) (ρ : Subst Γ Δ)
  : Substˢ ρ -> Subst⁻ Term ρ
substˢToSubst⁻ (∅ˢ Δ)        = ∅⁻ Term Δ
substˢToSubst⁻ (ρˢ ▷ˢ v)     = substˢToSubst⁻ ρˢ ▷⁻ valToTm⁻ v
substˢToSubst⁻ (ρˢ ∷≡⇒ˢ eqρ) = substˢToSubst⁻ ρˢ ∷≡⇒⁻ eqρ

abstract

  substˢToSubst⁻Eq (Γ, Δ : Ctxt) (ρ : Subst Γ Δ)
    : (ρˢ : Substˢ ρ) -> substˢToSubst⁻ ρˢ ≘⁻ substToSubst⁻ ρ
  substˢToSubst⁻Eq {Γ} {Δ} {ρ} ρˢ = substEqSubst⁻Eq $ reflSubst ρ

------------------------------------------------------------------------
-- Generalised value substitution casts

-- Basically a copy of ∷⇒⁻.

(∷⇒ˢ) (Γ¹, Δ¹ : Ctxt) (ρ¹ : Subst Γ¹ Δ¹)
      (Γ², Δ² : Ctxt) (ρ² : Subst Γ² Δ²)
  : Substˢ ρ¹ -> ρ¹ ≘ ρ² -> Substˢ ρ²
∅ˢ Δ ∷⇒ˢ emptyCong eqΔ = ∅ˢ _ ∷≡⇒ˢ symSubst (emptyCtxtSubstIs∅ _)
((▷ˢ) {_} {_} {_} {ρ¹} {t} ρˢ v) ∷⇒ˢ
  snocCong {_} {_} {_} {_} {_} {_} {_} {ρ²} eqρ eqτ eqt =
  (ρˢ ∷⇒ˢ lemma₁) ▷ˢ (v ∷ˢ lemma₂)
  ∷≡⇒ˢ symSubst (nonEmptyCtxtSubstIs▷ ρ²)
  where
  lemma₁ = symSubst (initIsInit _ ρ¹ t) `transSubst` eqρ
  lemma₂ = symTm    (lastIsLast _ ρ¹ t) `transTm`    last'EqLastEq eqt
(ρ ∷≡⇒ˢ eqρ₁) ∷⇒ˢ eqρ₂ = ρ ∷⇒ˢ eqρ₁ `transSubst` eqρ₂

------------------------------------------------------------------------
-- Looking up things from an environment

-- Basically a copy of /∋⁻.

(/∋ˢ) (Γ, Δ : Ctxt) (τ : Ty Γ) (ρ : Subst Γ Δ)
  : (v : Γ ∋ τ) -> Substˢ ρ -> Val (var v /⊢ ρ)
v           /∋ˢ (ρ ∷≡⇒ˢ eqρ) = v /∋ˢ ρ ∷ˢ substTmCong (reflTm _) eqρ
vz σ        /∋ˢ (ρ ▷ˢ val)   = val     ∷ˢ symTm (substVzSub' _ _ _)
vs v σ      /∋ˢ (ρ ▷ˢ val)   = v /∋ˢ ρ ∷ˢ symTm (substVsSub' _ _ _ _)
(v ∷≡∋ eqτ) /∋ˢ ρ            = v /∋ˢ ρ ∷ˢ symTm (substCastVar _ _ _)

------------------------------------------------------------------------
-- Various substitutions

subst⁻ToSubstˢ (Γ, Δ : Ctxt) (ρ : Subst Γ Δ)
  : (ρ⁻ : Subst⁻ Var ρ) -> Substˢ ρ
subst⁻ToSubstˢ (∅⁻ Var Δ) = ∅ˢ Δ
subst⁻ToSubstˢ (ρ⁻ ▷⁻ t⁻) =
  subst⁻ToSubstˢ ρ⁻ ▷ˢ (varToVal (tm⁻ToVar t⁻) ∷ˢ tm⁻ToVarEq t⁻)
subst⁻ToSubstˢ (ρ⁻ ∷≡⇒⁻ eqρ) = subst⁻ToSubstˢ ρ⁻ ∷≡⇒ˢ eqρ

idˢ : (Γ : Ctxt) -> Substˢ (id Γ)
idˢ Γ = subst⁻ToSubstˢ (id⁻ Var Γ)

wk⇒ˢ (Γ, Δ : Ctxt) (ρ : Subst Γ Δ)
  : Substˢ ρ -> (σ : Ty Δ) -> Substˢ (wk⇒ ρ σ)
wk⇒ˢ (∅ˢ Δ) σ = ∅ˢ (Δ ▻ σ) ∷≡⇒ˢ symSubst (emptyCtxtSubstIs∅ _)
wk⇒ˢ ((▷ˢ) {Γ} {Δ} {τ} {ρ} {t} ρˢ v) σ =
  wk⇒ˢ ρˢ σ ▷ˢ (wkVal v σ ∷≛ˢ eqτ) ∷≡⇒ˢ lemma
  where
  abstract
    eqτ : τ / ρ / wk σ ≛ τ / wk⇒ ρ σ
    eqτ = symTy (compSplitsTy _ _ _)

    lemma : wk⇒ ρ σ ▷ (t /⊢ wk σ ∷⊢ eqτ) ≘ wk⇒ (ρ ▷ t) σ
    lemma =
      Cong▷ (reflSubst _)
            (reflTy _)
            (symTm (generalCastTmEq _ _)
               `transTm`
             symTm (castTmEq _ _))
        `transSubst`
      symSubst (wkSnocIsSnocWk ρ τ t σ eqτ)
wk⇒ˢ (ρˢ ∷≡⇒ˢ eqρ) σ = wk⇒ˢ ρˢ σ ∷≡⇒ˢ wk⇒Cong eqρ (reflTy σ)

wk⇒ˢ⋆ (Γ, Δ : Ctxt) (ρ : Subst Γ Δ)
  : Substˢ ρ -> (Δ' : RCtxt Δ) -> Substˢ (wk⇒⋆ ρ Δ')
wk⇒ˢ⋆ ρ (εʳ _)    = ρ
wk⇒ˢ⋆ ρ (Δ' ▻ʳ σ) = wk⇒ˢ (wk⇒ˢ⋆ ρ Δ') σ

Normalisation-Environments : Type
Normalisation-Environments =
  time-plugin(Normalisation-Environments, Set)
