------------------------------------------------------------------------
-- Values
------------------------------------------------------------------------

-- # import Explicit
-- # import Implicit
-- # import Normalisation.NormalForms

-- The ˢ index stands for "_s_emantical value".

infixl 90 @ˢ
infixl 45 ∷ˢ, ∷≛ˢ

------------------------------------------------------------------------
-- Values

  -- This data type is a little problematic, since Val occurs
  -- negatively in ΠVal and ΠValCong. However, notice that the spines
  -- of the types of the recursive components are decreasing in
  -- ΠVal/ΠValCong, and non-increasing in all other constructors. This
  -- should ensure that the type is well-founded. I'm not aware of any
  -- meta-theory explaining a type such as Val, though.

mutual

  Val (Γ : Ctxt) (τ : Ty Γ) : Γ ⊢ τ -> Set
  Val = Val' _ _

  data Val' :  (Γ : Ctxt) -> (τ : Ty Γ) -> Γ ⊢ τ -> Set where
    ΠVal (Γ : Ctxt) (τ₁ : Ty Γ) (τ₂ : Ty (Γ ▻ τ₁)) (t₁ : Γ ⊢ Π τ₁ τ₂)
      : ΠFun t₁ -> Val t₁
    ElVal (Γ : Ctxt) (t' : Γ ⊢ ✶ Γ) (t : Γ ⊢ El t')
      : Atom t -> Val t
    ✶Val (Γ : Ctxt) (t : Γ ⊢ ✶ Γ)
      : Atom t -> Val t
    (∷ˢ) (Γ¹ : Ctxt) (τ¹ : Ty Γ¹) (t¹ : Γ¹ ⊢ τ¹)
         (Γ² : Ctxt) (τ² : Ty Γ²) (t² : Γ² ⊢ τ²)
      : Val t¹ -> t¹ ≈⊢ t² -> Val t²

  ΠFun (Γ : Ctxt) (τ₁ : Ty Γ) (τ₂ : Ty (Γ ▻ τ₁))
    : (t₁ : Γ ⊢ Π τ₁ τ₂) -> Set
  ΠFun {Γ} {τ₁} t₁ =
       (Γ' : RCtxt Γ)
    -> (t₂ : Γ ++ Γ' ⊢ τ₁ / wk⋆ Γ')
    -> (v₂ : Val t₂)
    -> Val ((t₁ /⊢ wk⋆ Γ') @ t₂)

------------------------------------------------------------------------
-- Possible alternative definition of Val

  -- Val could presumably be defined roughly as follows:

  -- Val (Γ : Ctxt) (τ : Ty Γ)
  --   : (s : Spine) -> (τs : TySp s τ) -> Γ ⊢ τ -> Set
  -- Val               ✶Sp         (✶TySp  Γ)      t  = Atom t
  -- Val               ElSp        (ElTySp t)      t  = Atom t
  -- Val {Γ} {Π τ₁ τ₂} (ΠSp s₁ s₂) (ΠTySp τs₁ τs₂) t₁ =
  --      (Γ' : RCtxt Γ)
  --   -> (t₂ : Γ ++ Γ' ⊢ τ₁ / wk⋆ Γ')
  --   -> (v₂ : Val s₁ (τs₁ /Sp _) t₂)
  --   -> Val s₂ (τs₂ /Sp _ /Sp _) ((t₁ /⊢ wk⋆ Γ') @ t₂)

  -- This way we can avoid the discussion about negative occurrences.
  -- I prefer to see this as an indication that the negative
  -- occurrences are not really a problem, though.

  -- Note that something more may have to be added to the alternative
  -- definition of Val in order to handle casts.

------------------------------------------------------------------------
-- Cast lemmas and specialised casts

(∷≛ˢ) (Γ¹ : Ctxt) (τ¹ : Ty Γ¹)
      (Γ² : Ctxt) (τ² : Ty Γ²)
      (t : Γ¹ ⊢ τ¹)
  : Val t -> (eqτ : τ¹ ≛ τ²) -> Val (t ∷⊢ eqτ)
v ∷≛ˢ eqτ = v ∷ˢ generalCastTmEq _ eqτ

------------------------------------------------------------------------
-- Weakening of values

wkVal (Γ : Ctxt) (τ : Ty Γ) (t : Γ ⊢ τ)
  : Val t -> (σ : Ty Γ) -> Val (t /⊢ wk σ)
wkVal (✶Val  at) σ = ✶Val  (at /⊨′ wk⁻ Var σ)
wkVal (ElVal at) σ = ElVal (at /⊨′ wk⁻ Var σ)
wkVal ((∷ˢ) {Γ¹} {_} {t¹} {Γ²} {_} {t²} v eqt) σ =
  wkVal v (σ ∷✶ eqΓ) ∷ˢ lemma
  where
  abstract
    eqΓ : Γ² ≜ Γ¹
    eqΓ = symCtxt $ tmEqCtxtEq eqt

    lemma : t¹ /⊢ wk (σ ∷✶ eqΓ) ≈⊢ t² /⊢ wk σ
    lemma = substTmCong eqt (wkCong $ symTy $ castTyEq σ eqΓ)
wkVal {Γ} {Π τ₁ τ₂} {t₁} (ΠVal f) σ = ΠVal f'
  where
  abstract
    lemma₁
      : (Γ' : RCtxt (Γ ▻ σ)) -> τ₁ / wk σ / wk⋆ Γ' ≛ τ₁ / wk⋆ (σ ◅ʳ Γ')
    lemma₁ Γ' = symTy $ wk⋆SplitsˡTy τ₁ Γ'

    lemma₂
      :  (Γ' : RCtxt (Γ ▻ σ))
      -> (t₂ : Γ ▻ σ ++ Γ' ⊢ τ₁ / wk σ / wk⋆ Γ')
      -> (t₁ /⊢ wk⋆ (σ ◅ʳ Γ')) @ (t₂ ∷⊢ lemma₁ Γ') ≈⊢
         (t₁ /⊢ wk σ /⊢ wk⋆ Γ') @ t₂
    lemma₂ Γ' t₂ =
      appCong
        (substTmCong (reflTm t₁) (wk⋆Splitsˡ Γ')
           `transTm`
         compSplitsTm t₁ (wk σ) (wk⋆ Γ'))
        (symTm $ generalCastTmEq t₂ _)

  f' : ΠFun (t₁ /⊢ wk σ)
  f' Γ' t₂ v₂ = f (σ ◅ʳ Γ') _ (v₂ ∷≛ˢ lemma₁ Γ')
                ∷ˢ lemma₂ Γ' t₂

wk⋆Val (Γ : Ctxt) (τ : Ty Γ) (t : Γ ⊢ τ)
  : Val t -> (Γ' : RCtxt Γ) -> Val (t /⊢ wk⋆ Γ')
wk⋆Val v (εʳ _)    = v ∷ˢ symTm (idVanishesTm _)
wk⋆Val v (Γ' ▻ʳ σ) =
  wkVal (wk⋆Val v Γ') σ ∷ˢ symTm (compSplitsTm _ _ _)

------------------------------------------------------------------------
-- Applying values

(@ˢ) (Γ : Ctxt) (τ₁ : Ty Γ) (τ₂ : Ty (Γ ▻ τ₁))
     (t₁ : Γ ⊢ Π τ₁ τ₂) (t₂ : Γ ⊢ τ₁)
  : Val t₁ -> Val t₂ -> Val (t₁ @ t₂)
(@ˢ) {Γ} {τ₁} {τ₂} {t₁} {t₂} (ΠVal f) v₂ =
  f (εʳ _) _ (v₂ ∷≛ˢ eqτ) ∷ˢ lemma
  where
  abstract
    eqτ : τ₁ ≛ τ₁ / wk⋆ (εʳ Γ)
    eqτ = symTy (idVanishesTy _)

    lemma : (t₁ /⊢ wk⋆ (εʳ Γ)) @ (t₂ ∷⊢ eqτ) ≈⊢ t₁ @ t₂
    lemma = appCong (idVanishesTm {_} {Π _ _} _)
                    (symTm $ generalCastTmEq _ _)
(v₁ ∷ˢ eqt) @ˢ v₂ with tmEqTyEq eqt =>
  _ (v₁ ∷ˢ eqt) v₂ (ΠCong eqτ₁ eqτ₂) =
    v₁ @ˢ (v₂ ∷≛ˢ symTy eqτ₁)
    ∷ˢ appCong eqt (symTm $ generalCastTmEq _ _)

------------------------------------------------------------------------
-- Reification and reflection

mutual

  reflect' (Γ : Ctxt) (τ : Ty Γ) (t : Γ ⊢ τ)
    : (s : Spine) -> TySp s τ -> Atom t -> Val t
  reflect' ✶Sp         (✶TySp _)       at = ✶Val  at
  reflect' ElSp        (ElTySp _)      at = ElVal at
  reflect' (ΠSp s₁ s₂) (ΠTySp τs₁ τs₂) at = ΠVal $ \Γ' _ v ->
    reflect' s₂ (τs₂ /Sp _ /Sp _)
             ((at /⊨′ wk⋆⁻ Var Γ') @⊨ reify' s₁ (τs₁ /Sp _) v)

  reify' (Γ : Ctxt) (τ : Ty Γ) (t : Γ ⊢ τ)
    : (s : Spine) -> TySp s τ -> Val t -> NF t
  reify' s τs (v ∷ˢ eqt) =
    reify' s (τs ∷Sp symTy (tmEqTyEq eqt)) v ∷⊨ eqt
  reify' _ _ (✶Val  at) = ✶Atom  at
  reify' _ _ (ElVal at) = ElAtom at
  reify' {Γ} {Π τ₁ τ₂} {t} (ΠSp s₁ s₂) (ΠTySp τs₁ τs₂) (ΠVal f) =
    λ⊨ (reify' s₂ (τs₂ /Sp _ /Sp _)
               (f (singletonʳ τ₁) _
                  (reflect' s₁ (τs₁ /Sp _) (var⊨ $ vz τ₁)
                   ∷≛ˢ lemma₁)))
    ∷≡⊨ lemma₂
    where
    abstract
      lemma₁ : τ₁ / wk τ₁ ≛ τ₁ / wk⋆ (singletonʳ τ₁)
      lemma₁ = substTyCong (reflTy τ₁) (symSubst $ singleWk⋆IsWk τ₁)

      lemma₂ : λ ((t /⊢ wk⋆ (singletonʳ τ₁)) @
                  (var (vz τ₁) ∷⊢ lemma₁)) ≈⊢
               t
      lemma₂ =
        λCong (appCong (substTmCong (reflTm t) (singleWk⋆IsWk τ₁))
                       (symTm $ generalCastTmEq _ _))
          `transTm`
        η t

reflect (Γ : Ctxt) (τ : Ty Γ) (t : Γ ⊢ τ)
  : Atom t -> Val t
reflect {Γ} {τ} at = reflect' (spine τ) (tyToTySp τ) at

reify (Γ : Ctxt) (τ : Ty Γ) (t : Γ ⊢ τ)
  : Val t -> NF t
reify {Γ} {τ} v = reify' (spine τ) (tyToTySp τ) v

------------------------------------------------------------------------
-- Some conversions to and from values

valToTm⁻ (Γ : Ctxt) (τ : Ty Γ) (t : Γ ⊢ τ)
  : Val t -> Tm⁻ Term t
valToTm⁻ v = nfToTm⁻ (reify v)

abstract

  valToTm⁻Eq (Γ : Ctxt) (τ : Ty Γ) (t : Γ ⊢ τ)
    : (v : Val t) -> valToTm⁻ v ≈⊢⁻ tmToTm⁻ t
  valToTm⁻Eq v = nfToTm⁻Eq (reify v)

varToVal (Γ : Ctxt) (τ : Ty Γ)
  : (v : Γ ∋ τ) -> Val (var v)
varToVal v = reflect (var⊨ v)

Normalisation-Values : Type
Normalisation-Values = time-plugin(Normalisation-Values, Set)
